@@ -992,3 +992,87 @@ class Nested_with_multilevel_query(Query):
     def TMT_nl(self) -> str:
         return """
             """
+
+class TestQuery2(Query):
+    @property
+    def sql(self):
+        return """SELECT m1.id, r1.stars FROM movie m1, rating r1 WHERE r1.stars < (SELECT q1.max_stars FROM q1result q1 WHERE q1.m_id = m1.id) AND m1.id IN (SELECT q2.id FROM q2result q2) GROUP BY m1.id)"""
+
+    @property
+    def graph(self):
+        raise NotImplementedError("TestQuery2 is not implemented yet")
+
+    @property
+    def simplified_graph(self):
+        if not TestQuery2._graph:
+            # Relation
+            movie = Relation("movie", "movie", is_primary = True)
+            Q1result = Relation("Q1result", "Q1's results")
+            Q2result = Relation("Q2result", "Q2's results")
+            rating = Relation("rating", "rating", is_primary = True)
+
+            # Attribute
+            movie_id1 = Attribute("m1_id1", "id")
+            movie_id2 = Attribute("m1_id2", "id")
+            movie_id3 = Attribute("m1_id3", "id")
+            movie_id4 = Attribute("m1_id4", "id")
+
+            rating_stars1 = Attribute("r1_stars", "stars")
+            rating_stars2 = Attribute("r1_stars1", "stars")
+
+            Q1_max_stars = Attribute("Q1_max_stars", "max_stars")
+            Q1_movie_id = Attribute("Q1_movie_id", "movie_id")
+
+            Q2_movie_ids = Attribute("Q2_movie_ids", "movie_id")
+            
+            rating_avg = Function(FunctionType.Avg)
+
+
+            ## Construct graph
+            query_graph = Query_graph("CorrelatedNestedQuery3")
+            query_graph.connect_membership(movie, movie_id1)
+            query_graph.connect_membership(rating, rating_stars1)
+            query_graph.connect_transformation(rating_avg, rating_stars1)
+
+            query_graph.connect_simplified_join(movie, rating, "", "belongs to")
+
+            # Nesting
+            query_graph.connect_selection(rating, rating_stars2)
+            query_graph.connect_predicate(rating_stars2, Q1_max_stars, OperatorType.LessThan)
+
+            query_graph.connect_membership(Q1result, Q1_max_stars)
+
+            # Correlation - skip
+            query_graph.connect_selection(Q1result, Q1_movie_id)
+            query_graph.connect_predicate(Q1_movie_id, movie_id4)
+            query_graph.connect_selection(movie_id4, movie)
+
+            # Second Nesting
+            query_graph.connect_selection(movie, movie_id2)
+            query_graph.connect_predicate(movie_id2, Q2_movie_ids, OperatorType.In)
+
+            query_graph.connect_membership(Q2result, Q2_movie_ids)
+            # For grouping and having
+            query_graph.connect_grouping(movie, movie_id3)
+            return query_graph
+            
+            Nested_with_multilevel_query._graph = query_graph
+        return TestQuery2._graph
+
+    @property
+    def nl(self):
+        return ""
+
+    @property
+    def BST_nl(self) -> str:
+        return """
+            """
+
+    @property
+    def MRP_nl(self) -> str:
+        return "Find id of movie for each id of movie where id of movie is in movie_id of Q2's results, average stars of rating in which rating belongs to movie and stars of rating is less than max_stars of Q1's results where movie_id of movie is id of those Q1's results."
+
+    @property
+    def TMT_nl(self) -> str:
+        return """
+            """
